#include <stdio.h>
#include <stdlib.h>
#include "util.h"
#include "bitstream.h"
#include "guessfsb.h"
#include "fsbext.h"
#include "xma_rebuild.h"

// XMAsh - decrypt, demux, and rebuild FSB XMA2
// cobbled together from romchu, guessfsb, decfsb, fsb_mpeg, and xma_parse

#define VERSION "0.0"

struct main_info {
    const char *file_name;
};

int try_key_callback(const uint8_t * d, long size, void *v);
int fsb_subfile_callback(const uint8_t * infile, long size, int streams, long samples, long srate, const char *stream_name, void *v);

int main(int argc, char **argv)
{
    long file_size;
    uint8_t *indata;

    int success = 0;

    struct main_info mi;

    if (argc != 2)
    {
        fprintf(stderr, "XMAsh " VERSION " - decrypt, demux, and rebuild FSB XMA2\n");
        fprintf(stderr, "usage: %s input.fsb.xen\n", argv[0]);
        exit(EXIT_FAILURE);
    }

    char * infile_name = argv[1];

    mi.file_name = strip_path(infile_name);

    {
        FILE *infile;
        infile = fopen(infile_name, "rb");
        CHECK_ERRNO(!infile, "open input failed");
        indata = get_whole_file(infile, &file_size);
        fclose(infile);
    }

    printf("%s\n\n", infile_name);

    // try without decrypting
    if (0 == try_multistream_fsb(indata, file_size, fsb_subfile_callback, &mi))
    {
        printf("success without decryption!\n");
        success = 1;
    }
    // try via guessfsb
    else if (0 == guess_fsb_keys(indata, file_size, try_key_callback, &mi))
    {
        printf("success!\n");
        success = 1;
    }
    else
    {
        printf("failure.\n");
    }

    free(indata);

    if (success)
    {
        return 0;
    }
    else
    {
        return 1;
    }
}

// called for each possibly valid key
int try_key_callback(const uint8_t * indata, long size, void *v)
{
    return try_multistream_fsb(indata, size, fsb_subfile_callback, v);
}

// called for each subfile in an FSB
int fsb_subfile_callback(const uint8_t * infile, long size, int streams, long samples, long srate, const char *stream_name, void *v)
{
    struct main_info *mip = v;

    // build the name
    int name_base_len = strlen(mip->file_name) + 1 + strlen(stream_name);
    char *name_base = malloc(name_base_len+1);
    CHECK_ERRNO(!name_base, "malloc");
    sprintf(name_base, "%s_%s", mip->file_name, stream_name);

    // rebuild each stream
    for (int str = 0; str < streams; str ++)
    {
        long parsed_samples;
        char *strname = number_name(name_base, ".xma", str+1, streams);
        printf("%s\n", strname);

        FILE *outfile = fopen(strname, "wb");
        CHECK_ERRNO(!outfile, "fopen output");
        CHECK_ERRNO(
            -1 == fseek(outfile, xma_header_size, SEEK_SET), "fseek past header");

        // rebuild!
        if (0 != build_XMA_from_XMA2(infile + packet_size_bytes*str,
                                     size-packet_size_bytes*str,
                                     outfile, 0x8000, 2, &parsed_samples))
        {
            // encountered an error while parsing
            fclose(outfile);
            free(strname);
            free(name_base);
            return 1;
        }

        parsed_samples += 128; // seems to always be short by 128
        if (parsed_samples != samples)
        {
            fprintf(stderr,"parsed samples=%ld, FSB said %ld\n", parsed_samples, samples);
            fclose(outfile);
            free(strname);
            free(name_base);
            return 1;
        }

        // write header
        long finish_offset = ftell(outfile);
        uint8_t *xma_head = make_xma_header(srate, finish_offset-xma_header_size, 2);
        CHECK_ERRNO(
            -1 == fseek(outfile, 0, SEEK_SET), "fseek to header");
        CHECK_ERRNO( 1 != fwrite(xma_head, xma_header_size, 1, outfile) , "fwrite header");

        CHECK_ERRNO(EOF == fclose(outfile), "fclose");
        free(xma_head);

        free(strname);
    }

    free(name_base);
    return 0;
}


